import base64
from enum import Enum
from .model import TypedModel
from .transcriber import TranscriberConfig
from .agent import AgentConfig
from .synthesizer import SynthesizerConfig

class WebSocketMessageType(str, Enum):
    BASE = 'base'
    START = 'start'
    AUDIO = 'audio'
    READY = 'ready'
    STOP = 'stop'

class WebSocketMessage(TypedModel, type=WebSocketMessageType.BASE): pass

class AudioMessage(WebSocketMessage, type=WebSocketMessageType.AUDIO):
    data: str

    @classmethod
    def from_bytes(cls, chunk: bytes):
        return cls(data=base64.b64encode(chunk).decode('utf-8'))

    def get_bytes(self) -> bytes:
        return base64.b64decode(self.data)

class StartMessage(WebSocketMessage, type=WebSocketMessageType.START):
    transcriber_config: TranscriberConfig
    agent_config: AgentConfig
    synthesizer_config: SynthesizerConfig

class ReadyMessage(WebSocketMessage, type=WebSocketMessageType.READY):
    pass

class StopMessage(WebSocketMessage, type=WebSocketMessageType.STOP):
    pass


from enum import Enum
from .model import TypedModel
from .audio_encoding import AudioEncoding
from ..output_device.base_output_device import BaseOutputDevice

class SynthesizerType(str, Enum):
    BASE = "base"
    AZURE = "azure"
    GOOGLE = "google"
    ELEVEN_LABS = "eleven_labs"

class SynthesizerConfig(TypedModel, type=SynthesizerType.BASE):
    sampling_rate: int
    audio_encoding: AudioEncoding

    @classmethod
    def from_output_device(cls, output_device: BaseOutputDevice):
        return cls(sampling_rate=output_device.sampling_rate, audio_encoding=output_device.audio_encoding)

class AzureSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.AZURE):
    pass

class GoogleSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.GOOGLE):
    pass

class ElevenLabsSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.ELEVEN_LABS):
    pass



from pydantic import BaseModel
from vocode.models.agent import AgentConfig, InformationRetrievalAgentConfig


class CallEntity(BaseModel):
    phone_number: str
    descriptor: str


class CreateCallRequest(BaseModel):
    recipient: CallEntity
    caller: CallEntity
    agent_config: InformationRetrievalAgentConfig  # TODO switch to AgentConfig
    # TODO add IVR/etc.



from typing import Optional
from enum import Enum
from .model import TypedModel, BaseModel


class AgentType(str, Enum):
    BASE = "base"
    LLM = "llm"
    CHAT_GPT_ALPHA = "chat_gpt_alpha"
    CHAT_GPT = "chat_gpt"
    ECHO = "echo"
    INFORMATION_RETRIEVAL = "information_retrieval"
    RESTFUL_USER_IMPLEMENTED = "restful_user_implemented"
    WEBSOCKET_USER_IMPLEMENTED = "websocket_user_implemented"


class AgentConfig(TypedModel, type=AgentType.BASE):
    initial_message: Optional[str] = None
    generate_responses: bool = True

class LLMAgentConfig(AgentConfig, type=AgentType.LLM):
    prompt_preamble: str
    expected_first_prompt: Optional[str] = None

class ChatGPTAlphaAgentConfig(AgentConfig, type=AgentType.CHAT_GPT_ALPHA):
    prompt_preamble: str
    expected_first_prompt: Optional[str] = None

class ChatGPTAgentConfig(AgentConfig, type=AgentType.CHAT_GPT):
    prompt_preamble: str
    expected_first_prompt: Optional[str] = None
    generate_responses: bool = False

class InformationRetrievalAgentConfig(
    AgentConfig, type=AgentType.INFORMATION_RETRIEVAL
):
    recipient_descriptor: str
    caller_descriptor: str
    goal_description: str
    fields: list[str]
    # TODO: add fields for IVR, voicemail


class EchoAgentConfig(AgentConfig, type=AgentType.ECHO):
    pass

class RESTfulUserImplementedAgentConfig(AgentConfig, type=AgentType.RESTFUL_USER_IMPLEMENTED):
    class EndpointConfig(BaseModel):
        url: str
        method: str = "POST"

    respond: EndpointConfig
    generate_responses: bool = False
    # generate_response: Optional[EndpointConfig]
    # update_last_bot_message_on_cut_off: Optional[EndpointConfig]

class RESTfulAgentInput(BaseModel):
    human_input: str

class RESTfulAgentOutput(BaseModel):
    response: str

class WebSocketUserImplementedAgentConfig(AgentConfig, type=AgentType.WEBSOCKET_USER_IMPLEMENTED):
    class RouteConfig(BaseModel):
        url: str

    respond: RouteConfig
    generate_responses: bool = False
    # generate_response: Optional[RouteConfig]
    # send_message_on_cut_off: bool = False

class WebSocketAgentMessageType(str, Enum):
    AGENT_BASE = 'agent_base'
    AGENT_START = 'agent_start'
    AGENT_TEXT = 'agent_text'
    AGENT_READY = 'agent_ready'
    AGENT_STOP = 'agent_stop'

class WebSocketAgentMessage(TypedModel, type=WebSocketAgentMessageType.AGENT_BASE): pass

class AgentTextMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.AGENT_TEXT):
    class Payload(BaseModel):
        text: str

    data: Payload

    @classmethod
    def from_text(cls, text: str):
        return cls(data=cls.Payload(text=text))


class AgentStartMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.AGENT_START):
    pass

class AgentReadyMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.AGENT_READY):
    pass

class AgentStopMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.AGENT_STOP):
    pass



from enum import Enum
from typing import Optional
from .audio_encoding import AudioEncoding
from .model import TypedModel
from ..input_device.base_input_device import BaseInputDevice

class TranscriberType(str, Enum):
    BASE = "base"
    DEEPGRAM = "deepgram"
    GOOGLE = "google"
    ASSEMBLY_AI = "assembly_ai"

class TranscriberConfig(TypedModel, type=TranscriberType.BASE):
    sampling_rate: int
    audio_encoding: AudioEncoding
    chunk_size: int

    @classmethod
    def from_input_device(cls, input_device: BaseInputDevice):
        return cls(
            sampling_rate=input_device.sampling_rate,
            audio_encoding=input_device.audio_encoding,
            chunk_size=input_device.chunk_size)

class DeepgramTranscriberConfig(TranscriberConfig, type=TranscriberType.DEEPGRAM):
    model: Optional[str] = None
    should_warmup_model: bool = False
    version: Optional[str] = None

class GoogleTranscriberConfig(TranscriberConfig, type=TranscriberType.GOOGLE):
    model: Optional[str] = None
    should_warmup_model: bool = False

class AssemblyAITranscriberConfig(TranscriberConfig, type=TranscriberType.ASSEMBLY_AI):
    should_warmup_model: bool = False


import websockets
import asyncio
from dotenv import load_dotenv
import os
import logging
import threading
import queue

load_dotenv()

from .input_device.base_input_device import BaseInputDevice
from .output_device.base_output_device import BaseOutputDevice
from .models.transcriber import TranscriberConfig
from .models.agent import AgentConfig
from .models.synthesizer import SynthesizerConfig
from .models.websocket import ReadyMessage, AudioMessage, StartMessage, StopMessage
from . import api_key

VOCODE_WEBSOCKET_URL = f'wss://api.vocode.dev/conversation'

class Conversation:

    def __init__(
        self,
        input_device: BaseInputDevice, 
        output_device: BaseOutputDevice, 
        transcriber_config: TranscriberConfig, 
        agent_config: AgentConfig,
        synthesizer_config: SynthesizerConfig
    ):
        self.input_device = input_device
        self.output_device = output_device
        self.transcriber_config = transcriber_config
        self.agent_config = agent_config
        self.synthesizer_config = synthesizer_config
        self.logger = logging.getLogger(__name__)
        self.receiver_ready = False
        self.active = True
        self.output_loop = asyncio.new_event_loop()
        self.output_audio_queue = queue.Queue()

    async def wait_for_ready(self):
        while not self.receiver_ready:
            await asyncio.sleep(0.1)
        return True
    
    def deactivate(self):
        self.active = False

    def play_audio(self):
        async def run():
            while self.active:
                try:
                    audio = self.output_audio_queue.get(timeout=5)
                    await self.output_device.send_async(audio)
                except queue.Empty:
                    continue
        loop = asyncio.new_event_loop()
        loop.run_until_complete(run())
    
    async def start(self):
        async with websockets.connect(f"{VOCODE_WEBSOCKET_URL}?key={api_key}") as ws:
            async def sender(ws):
                start_message = StartMessage(
                    transcriber_config=self.transcriber_config, 
                    agent_config=self.agent_config, 
                    synthesizer_config=self.synthesizer_config
                )
                await ws.send(start_message.json())
                await self.wait_for_ready()
                self.logger.info("Listening...press Ctrl+C to stop")
                while self.active:
                    data = self.input_device.get_audio()
                    if data:
                        await ws.send(AudioMessage.from_bytes(data).json())
                        await asyncio.sleep(0)
                await ws.send(StopMessage().json())

            async def receiver(ws):
                ReadyMessage.parse_raw(await ws.recv())
                self.receiver_ready = True
                async for msg in ws:
                    audio_message = AudioMessage.parse_raw(msg)
                    self.output_audio_queue.put_nowait(audio_message.get_bytes())


            output_thread = threading.Thread(target=self.play_audio)
            output_thread.start()
            return await asyncio.gather(sender(ws), receiver(ws))




from fastapi import FastAPI, APIRouter
import uvicorn

class BaseAgent():

    def __init__(self):
        self.app = FastAPI()

    async def respond(self, human_input) -> str:
        raise NotImplementedError
    
    def run(self, host="localhost", port=3000):
        uvicorn.run(self.app, host=host, port=port)


import sounddevice as sd
import numpy as np

from .base_output_device import BaseOutputDevice
from ..models.audio_encoding import AudioEncoding

class SpeakerOutput(BaseOutputDevice):

    DEFAULT_SAMPLING_RATE = 44100

    def __init__(self, device_info: dict, sampling_rate: int = None, audio_encoding: AudioEncoding = AudioEncoding.LINEAR16):
        self.device_info = device_info
        sampling_rate = sampling_rate or int(self.device_info.get('default_samplerate', self.DEFAULT_SAMPLING_RATE))
        super().__init__(sampling_rate, audio_encoding)
        self.stream = sd.OutputStream(
            channels=1,
            samplerate=self.sampling_rate,
            dtype=np.int16,
            device=int(self.device_info['index'])
        )
        self.stream.start()

    async def send_async(self, chunk):
        self.stream.write(np.frombuffer(chunk, dtype=np.int16))

    def terminate(self):
        self.stream.close()


from .base_agent import BaseAgent
from pydantic import BaseModel
import typing
from fastapi import APIRouter, WebSocket
from ..models.agent import AgentStartMessage, AgentReadyMessage, AgentTextMessage, WebSocketAgentMessage, WebSocketAgentMessageType

class WebSocketAgent(BaseAgent):
        
    def __init__(self):
        super().__init__()
        self.app.websocket("/respond")(self.respond_websocket)

    async def respond_websocket(self, websocket: WebSocket):
        await websocket.accept()
        AgentStartMessage.parse_obj(await websocket.receive_json())
        await websocket.send_text(AgentReadyMessage().json())
        while True:
            message = WebSocketAgentMessage.parse_obj(await websocket.receive_json())
            if message.type == WebSocketAgentMessageType.AGENT_STOP:
                break
            text_message = typing.cast(AgentTextMessage, message)
            response = await self.respond(text_message.data.text)
            await websocket.send_text(AgentTextMessage.from_text(response).json())
        await websocket.close()

